import networkx as nx

from diagv import misc


def dibull():
    return misc.digraph(
        {
            "0": "1",
            "1": "23",
            "2": "3",
            "3": "4",
        }
    )


def diline(n):
    return nx.DiGraph([(i, i + 1) for i in range(n - 1)])


def distar(n):
    if n < 0:
        return nx.freeze(nx.DiGraph([(i + 1, 0) for i in range(-n)]))
    return nx.DiGraph([(0, i + 1) for i in range(n)])


def ditutte_fragment():
    return misc.digraph(
        {
            "A": "BC",
            "B": "DE",
            "C": "EF",
            "D": "GH",
            "E": "I",
            "F": "JK",
            "G": "LM",
            "H": "IM",
            "I": "J",
            "J": "N",
            "K": "OP",
            "L": "OQ",
            "M": "N",
            "N": "O",
        }
    )


def ditutte():
    graph = nx.generators.tutte_graph()
    tree = nx.traversal.bfs_tree(graph, 0)
    order = list(nx.topological_sort(tree))
    return nx.DiGraph(
        [
            (head, tail) if order.index(head) < order.index(tail) else (tail, head)
            for (head, tail) in graph.edges
        ]
    )



"""Functionality to visualize apps"""

from __future__ import annotations

import itertools
from collections import Mapping
from typing import Callable, Iterable, Iterator, Optional, Sequence

import more_itertools
import networkx as nx
import numpy as np
from numpy.typing import NDArray

from diagv.typing_utils import AdjacencyListT, HashableT, NormalizedAdjacencyList


def text_art(
    digraph: nx.DiGraph,
    ordering: Optional[Sequence[HashableT]] = None,
    fmt: Callable[[HashableT], str] = str,
) -> str:
    try:
        nx.find_cycle(digraph)
    except nx.NetworkXNoCycle:
        pass
    else:
        NotImplementedError("Only DAGs are supported so far")

    if ordering is None:
        ordering = list(nx.topological_sort(digraph))

    for i, node in enumerate(ordering):
        predecessors = set(ordering[:i])
        conflicts = set(digraph.successors(node)) & predecessors
        if conflicts:
            raise ValueError(f"Ordering error at {node} for {conflicts}")

    cells = _cells(digraph, ordering)
    return "".join(_fmt_cells(cells, fmt))


def _nodes(graph: AdjacencyListT[HashableT]) -> Iterator[HashableT]:
    """Yield nodes in graph in reproducible order"""
    return more_itertools.unique_everseen(itertools.chain(graph, *graph.values()))


def _inverted(
    graph: AdjacencyListT[HashableT],
) -> NormalizedAdjacencyList[HashableT]:
    """Return graph with all arrows reverted"""
    result: NormalizedAdjacencyList[HashableT] = {node: set() for node in _nodes(graph)}
    for node in result:
        for adjacent in graph.get(node, []):
            result[adjacent].add(node)
    return result


def _above_has_successor_to_the_right(
    row: int,
    col: int,
    successor_lists: Mapping[int, Iterable[int]],
) -> Optional[int]:
    successors = itertools.chain.from_iterable(
        successor_lists.get(predecessor, []) for predecessor in range(row)
    )
    try:
        return col < max(successors)
    except ValueError:
        return False


class Token(str):
    ...


SECTION = Token("-")
INTERSECTION = Token("+")
OVERPASS = Token("|")
PADDING = Token(" ")
NOTHING = Token("")


def _cells(digraph: nx.DiGraph, order: Sequence[HashableT]) -> NDArray:
    node2position = {v: i for i, v in enumerate(order)}
    predecessor_lists = {
        node2position[node]: [
            node2position[direct_predecessor]
            for direct_predecessor in direct_predecessors
        ]
        for node, direct_predecessors in digraph.pred.items()
    }
    successor_lists = _inverted(predecessor_lists)
    num_row = len(order)
    num_col = 4 * len(order)
    result: NDArray = np.ndarray((num_row, num_col), dtype=object)
    for row in range(num_row):
        for col in range(num_col):
            result[row, col] = "*"
    for i, row_node in enumerate(order):
        successors = set(digraph.successors(row_node))
        row = i
        for j, col_node in enumerate(order):
            predecessors = set(digraph.predecessors(col_node))
            col = j * 4
            if j < i:
                if j:
                    result[row, col + 0] = PADDING
                    result[row, col + 1] = PADDING
                else:
                    result[row, col + 0] = NOTHING
                    result[row, col + 1] = NOTHING
                result[row, col + 2] = PADDING
                result[row, col + 3] = PADDING

            elif i == j:
                node = row_node
                assert node == col_node
                if predecessors:
                    result[row, col + 0] = INTERSECTION
                    result[row, col + 1] = SECTION
                elif j:
                    result[row, col + 0] = PADDING
                    result[row, col + 1] = PADDING
                else:
                    result[row, col + 0] = NOTHING
                    result[row, col + 1] = NOTHING

                result[row, col + 2] = node

                if successors:
                    result[row, col + 3] = SECTION
                elif _above_has_successor_to_the_right(i, j, successor_lists):
                    result[row, col + 3] = PADDING
                else:
                    result[row, col + 3] = NOTHING
            else:
                if row_node in predecessors:
                    result[row, col + 0] = INTERSECTION
                elif (
                    predecessors
                    and min(map(node2position.__getitem__, predecessors)) < i
                ):
                    result[row, col + 0] = OVERPASS
                elif successors and j < max(map(node2position.__getitem__, successors)):
                    result[row, col + 0] = SECTION
                elif _above_has_successor_to_the_right(i, j, successor_lists):
                    result[row, col + 0] = PADDING
                else:
                    result[row, col + 0] = NOTHING

                if successors and j < max(map(node2position.__getitem__, successors)):
                    result[row, col + 1] = SECTION
                    result[row, col + 2] = SECTION
                    result[row, col + 3] = SECTION
                elif _above_has_successor_to_the_right(i, j, successor_lists):
                    result[row, col + 1] = PADDING
                    result[row, col + 2] = PADDING
                    result[row, col + 3] = PADDING
                else:
                    result[row, col + 1] = NOTHING
                    result[row, col + 2] = NOTHING
                    result[row, col + 3] = NOTHING

    return result


def _fmt_cells(cells: NDArray, fmt: Callable[[HashableT], str]) -> Iterator[str]:
    num_row, num_col = cells.shape
    col2width = {
        4 * row + 2: len(fmt(cells[row, 4 * row + 2])) for row in range(num_row)
    }
    for row in range(num_row):
        if row:
            yield "\n"
        for col in range(num_col):
            cell = cells[row, col]
            width = col2width.get(col, 1)
            if cell is NOTHING:
                yield ""
            elif cell in {PADDING, SECTION}:
                yield cell * width
            elif cell in {INTERSECTION, OVERPASS}:
                assert width == 1
                yield cell
            else:
                yield f"{fmt(cell)[:width]:.^{width}}"



import networkx as nx

from diagv.typing_utils import AdjacencyListT


def digraph(graph: AdjacencyListT, order=1) -> nx.DiGraph:
    return nx.DiGraph((k, v)[::order] for k, vs in graph.items() for v in vs)


def raise_for_cyclic(digraph: nx.DiGraph) -> None:
    try:
        nx.find_cycle(digraph)
    except nx.NetworkXNoCycle:
        return
    NotImplementedError("Only DAGs are supported so far")



from typing import Dict, Hashable, Iterable, Mapping, Set, TypeVar

HashableT = TypeVar("HashableT", bound=Hashable)
AdjacencyListT = Mapping[HashableT, Iterable[HashableT]]
NormalizedAdjacencyList = Dict[HashableT, Set[HashableT]]



import pathlib
from typing import Mapping, Sequence

import setuptools

DISTRIBUTION_PACKAGE_NAME = INSTALL_PACKAGE_NAME = "diagv"

FILE = pathlib.Path(__file__)
REQUIREMENTS = FILE.parent / "requirements"


def read_description() -> str:
    title = read_readme().splitlines()[0]
    prefix = "# "
    assert title.startswith(prefix)
    return title[len(prefix) :]


def read_readme() -> str:
    file = FILE.parent / "README.md"
    return file.read_text()


def read_requirements(name) -> Sequence[str]:
    file = REQUIREMENTS / f"{name}.txt"
    with file.open() as f:
        return list(f)


def read_install_requires() -> Sequence[str]:
    return read_requirements("install_requires")


def read_extras_require() -> Mapping[str, Sequence[str]]:
    return {
        file.stem.split("-", maxsplit=1)[1]: read_requirements(file.stem)
        for file in REQUIREMENTS.glob("extras_require-*.txt")
    }


setuptools.setup(
    author="AP Ljungquist",
    author_email="ap@ljungquist.eu",
    description=read_description(),
    extras_require=read_extras_require(),
    install_requires=read_install_requires(),
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    name=DISTRIBUTION_PACKAGE_NAME,
    package_data={
        INSTALL_PACKAGE_NAME: ["py.typed"],
    },
    package_dir={"": "src"},
    packages=setuptools.find_packages("src"),
    python_requires=">=3.9",
    url="https://github.com/apljungquist/diagv",
)



"""Functionality to visualize apps"""
from __future__ import annotations

import collections
import functools
import itertools
import math
import operator
from typing import Callable, Dict, FrozenSet, Generic, Iterator, Sequence, Set, Tuple

import more_itertools
import networkx as nx

from diagv import misc
from diagv.typing_utils import AdjacencyListT, HashableT, NormalizedAdjacencyList


def _nodes(graph: AdjacencyListT[HashableT]) -> Iterator[HashableT]:
    """Yield nodes in graph in reproducible order"""
    return more_itertools.unique_everseen(itertools.chain(graph, *graph.values()))


def _normalized(
    graph: AdjacencyListT[HashableT],
) -> NormalizedAdjacencyList[HashableT]:
    """Return graph in form that is amenable to many computations"""
    return {node: set(graph.get(node, [])) for node in _nodes(graph)}


def _inverted(
    graph: AdjacencyListT[HashableT],
) -> NormalizedAdjacencyList[HashableT]:
    """Return graph with all arrows reverted"""
    result: NormalizedAdjacencyList[HashableT] = {node: set() for node in _nodes(graph)}
    for node in result:
        for adjacent in graph.get(node, []):
            result[adjacent].add(node)
    return result


def _reachables(
    graph: AdjacencyListT[HashableT], origin: HashableT
) -> Iterator[HashableT]:
    """Yield all nodes that are reachable from `origin`"""
    done: Set[HashableT] = set()
    remaining = collections.deque(graph.get(origin, []))
    while remaining:
        node = remaining.popleft()
        yield node
        if node not in done:
            done.add(node)
            remaining.extend(graph.get(node, []))


def _reachability(
    graph: AdjacencyListT[HashableT],
) -> Dict[HashableT, Set[HashableT]]:
    """Return mapping from each node to all its descendants"""
    return {node: set(_reachables(graph, node)) for node in graph}


class _Graph(Generic[HashableT]):
    """Precomputed collection of graph properties

    These are frequently accessed by the algorithms and having them precomputed speeds
    up execution.
    """

    def __init__(self, graph: AdjacencyListT[HashableT]) -> None:
        self.nodes = list(
            more_itertools.unique_everseen(itertools.chain(graph, *graph.values()))
        )
        self.nodes_set = frozenset(self.nodes)
        self.node2direct_predecessors = _normalized(graph)
        self.node2direct_successors = _inverted(graph)
        self.node2predecessors = _reachability(self.node2direct_predecessors)
        self.node2successors = _reachability(self.node2direct_successors)


def marginal_cost(
    graph: _Graph[HashableT],
    before: Sequence[HashableT],
    this: HashableT,
    after: FrozenSet[HashableT],
) -> int:
    node2position = {v: i for i, v in enumerate(before)}
    dpreds = graph.node2direct_predecessors[this]
    dpred_positions = [
        node2position[direct_predecessor] for direct_predecessor in dpreds
    ]
    first_dpred_pos = min(dpred_positions, default=len(before))
    intermediate = [
        (
            len(graph.node2direct_successors[other] & after),
            (first_dpred_pos < other_pos and other not in dpreds),
        )
        # Before is often longer than after since invocations with short before get
        # cached. Is it possible to iterate over after instead? Equivalent complexity
        # but should get rid of one python constant
        for other_pos, other in enumerate(before)
    ]
    cost_from_other = sum(
        [
            num_extension + bool(num_extension) * extension_is_interaction
            for num_extension, extension_is_interaction in intermediate
        ]
    )
    return cost_from_other


@functools.lru_cache(maxsize=100_000)
def cost(
    graph: _Graph[HashableT],
    prefix: Sequence[HashableT],
    unvisited: FrozenSet[HashableT],
) -> int:
    """Returns a number quantifying how not nice the graph would look when drawn

    The returned number is a lower bound on the cost of any path that starts with `prefix`.
    I.e. there is no way to add the unvisited nodes to the prefix and get a lower cost.
    A tighter bound allows more effective pruning.
    """
    if not prefix:
        return 0
    before = prefix[:-1]
    this = prefix[-1]
    return cost(graph, before, unvisited | {this}) + marginal_cost(
        graph, before, this, unvisited
    )


def _topological_orderings(
    g,
    path: Tuple[HashableT, ...],
    should_be_pruned: Callable[[Tuple[HashableT, ...], FrozenSet[HashableT]], bool],
) -> Iterator[Tuple[HashableT, ...]]:
    """Yield topological orderings of graph"""
    visited = set(path)
    for node in g.nodes:
        if node in visited:
            continue

        # Prune branches that would not be ordered
        if g.node2predecessors[node] - visited:
            continue

        new_path = path + (node,)
        new_unvisited = g.nodes_set - visited - {node}
        # Prune branches that are guaranteed to be suboptimal
        if should_be_pruned(new_path, new_unvisited):
            continue

        if new_unvisited:
            yield from _topological_orderings(g, new_path, should_be_pruned)
        else:
            yield new_path


def sorted_topological(digraph: nx.DiGraph) -> Tuple[HashableT, ...]:
    """Return an optimal topological ordering

    Not that there may be several others that are equally good.
    """
    misc.raise_for_cyclic(digraph)
    return min(as_found(digraph), key=operator.itemgetter(0))[1]


def as_found(digraph: nx.DiGraph) -> Iterator[Tuple[int, Tuple[HashableT, ...]]]:
    """Return good topological orderings as they are found

    Each ordering is no worse than the previous.
    """
    misc.raise_for_cyclic(digraph)
    g = _Graph(digraph.pred)
    prev_best = math.inf

    def prune(new_path, new_unvisited):
        return prev_best < cost(g, new_path, new_unvisited)

    for order in _topological_orderings(g, (), prune):
        curr_best = cost(g, order, frozenset())
        yield curr_best, order
        assert curr_best <= prev_best
        prev_best = curr_best
